# Trying again
import math


class Expression:
    def __add__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            other = Constant(other)
        return Addition(self, other)

    def __radd__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            other = Constant(other)
        return Addition(other, self)

    def __truediv__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError("Cannot divide by zero.")
            else:
                other = Constant(other)
        return Divide(self, other)

    def __rtruediv__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError("Cannot divide by zero.")
            else:
                other = Constant(other)
        return Divide(other, self)

    def __rmul__(self, other):
        return self * other

    def __mul__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            other = Constant(other)
        return Multiply(self, other)

    def __sub__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            other = Constant(other)
        return Subtraction(self, other)

    def __rsub__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            other = Constant(other)
        return Subtraction(other, self)

    def __neg__(self):
        return -1 * self

    def simplify(self):
        return self


class Constant(Expression):
    def __init__(self, val):
        if isinstance(val, int) or isinstance(val, float):
            self.val = val
        else:
            raise TypeError

    def __str__(self):
        return f"{self.val}"

    def __eq__(self, other):
        if isinstance(other, Constant):
            return self.val == other.val
        return False

    def evaluate(self, env):
        return self.val

    def derivative(self, name):
        return Constant(0)

    def simplify(self):
        return self

    def decompose(self):
        return []


class Variable(Expression):
    def __init__(self, name):
        if name[0].isalpha() and name[0].isascii():
            self.name = name
        else:
            raise TypeError

    def __str__(self):
        return f"{self.name}"

    def __eq__(self, other):
        if isinstance(other, Variable):
            return self.name == other.name
        return False

    # def __neg__(self):
    #     pass

    def evaluate(self, env):
        return Constant(env[self.name]).evaluate(env)

    def derivative(self, name):
        return Constant(1 if self.name == name else 0)

    def decompose(self):
        return [self.name]


class BinaryOp(Expression):
    def __init__(self, left, right):
        self.left = left
        self.right = right
        self.name = ""

    def __str__(self):
        return f"({self.left} {self.name} {self.right})"

    def evaluate(self, env):
        return self.operate(self.left.evaluate(env), self.right.evaluate(env))

    def decompose(self):
        return self.left.decompose() + self.right.decompose()


class Addition(BinaryOp):
    def __init__(self, left, right):
        super().__init__(left, right)
        self.name = "+"

    def operate(self, x, y):
        return x + y

    def derivative(self, name):
        return self.left.derivative(name) + self.right.derivative(name)

    def simplify(self):
        left = self.left.simplify()
        right = self.right.simplify()
        # if left.decompose() == []:
        #     left = left.evaluate({})
        # if right.decompose() == []:
        #     right = right.evaluate({})

        if isinstance(left, Constant) and isinstance(right, Constant):
            return Constant(left.val + right.val)

        if left == Constant(0):
            return right

        if right == Constant(0):
            return left

        return left + right


class Subtraction(BinaryOp):
    def __init__(self, left, right):
        super().__init__(left, right)
        self.name = "-"

    def operate(self, x, y):
        return x - y

    def derivative(self, name):
        return self.left.derivative(name) - self.right.derivative(name)

    def simplify(self):
        if self.left == self.right:
            return 0
        left = self.left.simplify()
        right = self.right.simplify()
        # if left.decompose() == []:
        #     left = left.evaluate({})
        # if right.decompose() == []:
        #     right = right.evaluate({})

        if isinstance(left, Constant) and isinstance(right, Constant):
            return Constant(left.val - right.val)

        if left == Constant(0):
            return -1 * right

        if right == Constant(0):
            return left

        return left - right


class Multiply(BinaryOp):
    def __init__(self, left, right):
        super().__init__(left, right)
        self.name = "*"

    def operate(self, x, y):
        return x * y

    def derivative(self, name):
        f = self.left
        f1 = f.derivative(name)
        g = self.right
        g1 = g.derivative(name)

        return f1 * g + f * g1

    def simplify(self):
        if self.left == Constant(0) or self.right == Constant(0):
            return Constant(0)

        if self.left == Constant(1):
            return self.right

        if self.right == Constant(1):
            return self.left
        left = self.left.simplify()
        right = self.right.simplify()
        # if left.decompose() == []:
        #     left = left.evaluate({})
        # if right.decompose() == []:
        #     right = right.evaluate({})

        if isinstance(left, Constant) and isinstance(right, Constant):
            return Constant(left.val * right.val)

        return left * right


class Divide(BinaryOp):
    def __init__(self, left, right):
        super().__init__(left, right)
        self.name = "/"

    def operate(self, x, y):
        return x / y

    def derivative(self, name):
        g = self.right
        g1 = g.derivative(name)
        f = self.left
        f1 = f.derivative(name)

        return ((g * f1) - f * g1) / (g * g)

    def simplify(self):
        if self.left == self.right:
            return 1
        left = self.left.simplify()
        right = self.right.simplify()
        # if left.decompose() == []:
        #     left = left.evaluate({})
        # if right.decompose() == []:
        #     right = right.evaluate({})

        if isinstance(left, Constant) and isinstance(right, Constant):
            return Constant(left.val - right.val)

        if left == Constant(0):
            return 0

        if right == Constant(1):
            return left

        return left / right


class UnaryOp(Expression):
    def __init__(self, power):
        self.power = power

    def evaluate(self, env):
        if isinstance(self.power, int) or isinstance(self.power, float):
            return self.operate(self.power)
        return self.operate(self.power.evaluate(env))


class Exponent(UnaryOp):

    def __str__(self):
        return f"e^{self.power}"

    def __eq__(self, other):
        if isinstance(other, Exponent):
            return self.power == other.power
        return False

    def operate(self, x):
        return math.exp(x)

    def derivative(self, name):
        return self.power.derivative(name) * Exponent(self.power)

    def decompose(self):
        return self.power

    def simplify(self):
        power = self.power.simplify()
        # if left.decompose() == []:
        #     left = left.evaluate({})
        # if right.decompose() == []:
        #     right = right.evaluate({})

        if isinstance(power, Constant):
            return self.operate(power)

        return Exponent(power)
